@@ -42,8 +42,8 @@ typedef struct _Parser
 	Py_UNICODE *stringparse_buffer;
 	size_t stringparse_buffer_size;
 	
-	unsigned int use_float: 1;
-	unsigned int got_root: 1;
+	unsigned char use_float;
+	unsigned char got_root;
 } Parser;
 
 static PyObject *
@@ -108,6 +108,7 @@ jsonlib_read (PyObject *self, PyObject *args)
 	parser.use_float = use_float;
 	parser.module = PyModule_GetState (self);
 	
+	skip_whitespace (&parser, NULL, NULL);
 	result = parse_raw (&parser);
 	skip_whitespace (&parser, NULL, NULL);
 	
@@ -126,7 +127,6 @@ parse_raw (Parser *parser)
 {
 	Py_UNICODE c;
 	
-	skip_whitespace (parser, NULL, NULL);
 	if (parser->index == parser->end)
 	{
 		return PyObject_CallMethod (parser->error_helper, "no_expression", "uk",
@@ -170,9 +170,8 @@ parse_raw (Parser *parser)
 	case '9':
 		return parse_number (parser);
 	default:
-		break;
+		return parser_raise_unexpected (parser, NULL);
 	}
-	return parser_raise_unexpected (parser, NULL);
 }
 
 static PyObject *
@@ -182,8 +181,7 @@ parse_object (Parser *parser)
 	Py_UNICODE *start = parser->index;
 	int result;
 	
-	if (!skip_char (parser, '{', "object start"))
-	{ goto error; }
+	++parser->index; /* '{' */
 	
 	if (!skip_whitespace (parser, start, "Unterminated object."))
 	{ goto error; }
@@ -191,9 +189,7 @@ parse_object (Parser *parser)
 	retval = PyDict_New ();
 	if (*parser->index == '}')
 	{
-		if (!skip_char (parser, '}', "object end"))
-		{ goto error; }
-		
+		++parser->index;
 		return retval;
 	}
 	
@@ -236,9 +232,7 @@ parse_object (Parser *parser)
 		
 		if (*parser->index == '}')
 		{
-			if (!skip_char (parser, '}', "object end"))
-			{ goto error; }
-			
+			++parser->index;
 			return retval;
 		}
 		
@@ -261,8 +255,7 @@ parse_array (Parser *parser)
 	Py_UNICODE *start = parser->index;
 	int result;
 	
-	if (!skip_char (parser, '[', "array start"))
-	{ goto error; }
+	++parser->index; /* '[' */
 	
 	if (!skip_whitespace (parser, start, "Unterminated array."))
 	{ goto error; }
@@ -270,9 +263,7 @@ parse_array (Parser *parser)
 	retval = PyList_New (0);
 	if (*parser->index == ']')
 	{
-		if (!skip_char (parser, ']', "object end"))
-		{ goto error; }
-		
+		++parser->index;
 		return retval;
 	}
 	
@@ -295,9 +286,7 @@ parse_array (Parser *parser)
 		
 		if (*parser->index == ']')
 		{
-			if (!skip_char (parser, ']', "array end"))
-			{ goto error; }
-			
+			++parser->index;
 			return retval;
 		}
 		
@@ -316,8 +305,8 @@ static PyObject *
 parse_string (Parser *parser)
 {
 	PyObject *unicode;
-	int escaped = 0, fancy = 0;
-	Py_UNICODE c, *start;
+	unsigned char escaped = 0, fancy = 0;
+	Py_UNICODE c, *start, *end = parser->end;
 	size_t ii;
 	
 	start = parser->index;
@@ -332,7 +321,7 @@ parse_string (Parser *parser)
 	/* Scan through for maximum character count, and to ensure the string
 	 * is terminated.
 	**/
-	for (ii = 1; start + ii < parser->end; ii++)
+	for (ii = 1; start + ii < end; ++ii)
 	{
 		c = start[ii];
 		
@@ -347,19 +336,17 @@ parse_string (Parser *parser)
 		if (escaped)
 		{ escaped = 0; }
 		
-		else
+		else if (c == '\\')
 		{
-			if (c == '\\')
-			{
-				fancy = 1;
-				escaped = 1;
-			}
-			else if (c == '"')
-			{ break; }
+			fancy = 1;
+			escaped = 1;
 		}
+		
+		else if (c == '"')
+		{ break; }
 	}
 	
-	if (start + ii >= parser->end)
+	if (start + ii >= end)
 	{ return parser_raise_unterminated_string (parser, start); }
 	
 	if (fancy)
@@ -421,7 +408,7 @@ parse_string_full (Parser *parser, Py_UNICODE *start, size_t max_char_count)
 				return unicode;
 			}
 			else { buffer[buffer_idx++] = c; }
-			ii++;
+			++ii;
 		}
 		
 		escaped = 0;
@@ -446,8 +433,8 @@ parse_string_full (Parser *parser, Py_UNICODE *start, size_t max_char_count)
 				parser, start, buffer, &buffer_idx, &ii,
 				max_char_count))
 			{
-				ii--;
-				buffer_idx++;
+				--ii;
+				++buffer_idx;
 				break;
 			}
 			else { return NULL; }
@@ -459,7 +446,7 @@ parse_string_full (Parser *parser, Py_UNICODE *start, size_t max_char_count)
 				(start - parser->start + ii - 1),
 				&c, 1);
 		}
-		ii++;
+		++ii;
 	}
 }
 
@@ -484,7 +471,7 @@ parse_unicode_escape (Parser *parser, Py_UNICODE *string_start, Py_UNICODE *buff
 	size_t offset = *index_ptr;
 	Py_UNICODE value;
 	
-	offset++;
+	++offset;
 	
 	remaining = max_char_count - offset;
 	if (remaining < 4)
@@ -565,21 +552,21 @@ static PyObject *
 keyword_compare (Parser *parser, const char *expected, size_t len,
                  PyObject *retval)
 {
-	size_t ii, left;
+	size_t ii;
+	Py_UNICODE *index = parser->index;
 	
-	left = parser->end - parser->index;
-	if (left >= len)
+	if (parser->end - index < len)
+	{ return NULL; }
+	
+	for (ii = 0; expected[ii]; ++ii)
 	{
-		for (ii = 0; ii < len; ii++)
-		{
-			if (parser->index[ii] != (unsigned char)(expected[ii]))
-			{ return NULL; }
-		}
-		parser->index += len;
-		Py_INCREF (retval);
-		return retval;
+		if (parser->index[ii] != (unsigned char)(expected[ii]))
+		{ return NULL; }
 	}
-	return NULL;
+	
+	parser->index += ii;
+	Py_INCREF (retval);
+	return retval;
 }
 
 static PyObject *
@@ -648,7 +635,7 @@ parse_number (Parser *parser)
 		}
 		if (should_stop)
 		{ break; }
-		ptr++;
+		++ptr;
 	}
 	
 	if (got_digit)
@@ -688,25 +675,30 @@ parse_number (Parser *parser)
 static unsigned char
 skip_whitespace (Parser *parser, Py_UNICODE *start, const char *message)
 {
-	Py_UNICODE c;
+	Py_UNICODE *index = parser->index, *end = parser->end;
 	
 	if (message && !start)
-	{ start = parser->index; }
+	{ start = index; }
 	
 	/* Don't use Py_UNICODE_ISSPACE, because it returns TRUE for
 	 * codepoints that are not valid JSON whitespace.
 	**/
-	while (parser->index < parser->end)
+	while (index < end)
 	{
-		c = *parser->index;
-		if (!(c == '\x09' ||
-		      c == '\x0A' ||
-		      c == '\x0D' ||
-		      c == '\x20'))
-		{ return 1; }
-		
-		parser->index++;
+		switch (*index)
+		{
+		case '\x09':
+		case '\x0A':
+		case '\x0D':
+		case '\x20':
+			++index;
+			break;
+		default:
+			parser->index = index;
+			return 1;
+		}
 	}
+	parser->index = index;
 	
 	if (message)
 	{
@@ -727,7 +719,7 @@ skip_char (Parser *parser, Py_UNICODE c, const char *message)
 		return 0;
 	}
 	
-	parser->index++;
+	++parser->index;
 	return 1;
 }
 
@@ -768,10 +760,11 @@ typedef struct _Serializer Serializer;
 struct _Serializer
 {
 	ModuleState *module;
+	Py_UNICODE *buffer;
+	size_t buffer_size;
 	
 	/* Virtual implementation methods */
-	unsigned char (*append_ascii) (Serializer *, const char *);
-	unsigned char (*append_unicode) (Serializer *, PyObject *);
+	unsigned char (*flush) (Serializer *);
 	
 	/* Per-run constant. This isn't saved in the module state
 	 * because it varies based on the indentation mode.
@@ -782,32 +775,36 @@ struct _Serializer
 	PyObject *indent;
 	PyObject *on_unknown;
 	PyObject *error_helper;
-	unsigned int sort_keys: 1;
-	unsigned int coerce_keys: 1;
-	unsigned int ascii_only: 1;
+	unsigned char sort_keys;
+	unsigned char coerce_keys;
+	unsigned char ascii_only;
+};
+
+typedef struct _BufferList BufferList;
+struct _BufferList
+{
+	Py_UNICODE *buffer;
+	size_t buffer_size;
+	BufferList *next;
 };
 
 typedef struct _BufferSerializer
 {
 	Serializer base;
-	Py_UNICODE *buffer;
-	size_t buffer_size;
-	size_t buffer_max_size;
+	const char *encoding;
+	BufferList *all_buffers;
+	BufferList *current_buffer;
 } BufferSerializer;
 
 typedef struct _StreamSerializer
 {
 	Serializer base;
 	PyObject *stream;
-	Py_UNICODE *buffer;
-	size_t buffer_size;
 	char *encoding;
-	unsigned int ascii_safe_encoding: 1;
 } StreamSerializer;
 
 static const char hexdigit[] = "0123456789abcdef";
-static const size_t INITIAL_BUFFER_SIZE = 32;
-static const size_t STREAM_BUFFER_SIZE = 1024;
+#define FIXED_BUFFER_SIZE 1024
 
 static PyObject *
 jsonlib_write (PyObject *, PyObject *);
@@ -863,22 +860,25 @@ static unsigned char
 serializer_raise_reserved_cp (Serializer *, Py_UNICODE);
 
 static unsigned char
-buffer_serializer_append_ascii (Serializer *, const char *);
+serializer_append_ascii (Serializer *, const char *);
 
 static unsigned char
-buffer_serializer_append_unicode (Serializer *, PyObject *);
+serializer_append_unicode (Serializer *, PyObject *);
 
 static unsigned char
-buffer_serializer_resize (BufferSerializer *, size_t delta);
+buffer_serializer_flush (Serializer *);
+
+static BufferList *
+alloc_buffer_list (void);
+
+static PyObject *
+buffer_serializer_merge (BufferSerializer *);
+
+static void
+buffer_serializer_free (BufferSerializer *);
 
 static unsigned char
-stream_serializer_append_ascii (Serializer *, const char *);
-
-static unsigned char
-stream_serializer_append_unicode (Serializer *, PyObject *);
-
-static unsigned char
-stream_serializer_flush (StreamSerializer *);
+stream_serializer_flush (Serializer *);
 
 static PyObject *
 ascii_constant (const char *value)
@@ -895,7 +895,6 @@ jsonlib_write (PyObject *self, PyObject *args)
 	
 	/* Parameters */
 	PyObject *value;
-	char *encoding;
 	unsigned char sort_keys, ascii_only, coerce_keys;
 	
 	if (!PyArg_ParseTuple(args, "ObObbzOO",
@@ -904,7 +903,7 @@ jsonlib_write (PyObject *self, PyObject *args)
 		&base->indent,
 		&ascii_only,
 		&coerce_keys,
-		&encoding,
+		&serializer.encoding,
 		&base->on_unknown,
 		&base->error_helper))
 	{ return NULL; }
@@ -915,27 +914,16 @@ jsonlib_write (PyObject *self, PyObject *args)
 	
 	/* Implementation pointers */
 	base->module = PyModule_GetState (self);
-	base->append_ascii = buffer_serializer_append_ascii;
-	base->append_unicode = buffer_serializer_append_unicode;
+	base->flush = buffer_serializer_flush;
 	
 	if (!serializer_run (base, value))
 	{ goto error; }
 	
-	if (encoding)
-	{
-		result = PyUnicode_Encode (
-			serializer.buffer, serializer.buffer_size,
-			encoding, "strict");
-	}
-	
-	else
-	{
-		result = PyUnicode_FromUnicode (
-			serializer.buffer, serializer.buffer_size);
-	}
+	/* Merge all buffers into a single unicode string */
+	result = buffer_serializer_merge (&serializer);
 	
 error:
-	PyMem_Free (serializer.buffer);
+	buffer_serializer_free (&serializer);
 	return result;
 }
 
@@ -947,9 +935,9 @@ jsonlib_dump (PyObject *self, PyObject *args)
 	
 	/* Parameters */
 	PyObject *value;
-	unsigned char sort_keys, ascii_only, coerce_keys, ascii_safe;
+	unsigned char sort_keys, ascii_only, coerce_keys;
 	
-	if (!PyArg_ParseTuple(args, "OObObbzbOO",
+	if (!PyArg_ParseTuple(args, "OObObbzOO",
 		&value,
 		&serializer.stream,
 		&sort_keys,
@@ -957,7 +945,6 @@ jsonlib_dump (PyObject *self, PyObject *args)
 		&ascii_only,
 		&coerce_keys,
 		&serializer.encoding,
-		&ascii_safe,
 		&base->on_unknown,
 		&base->error_helper))
 	{ return NULL; }
@@ -966,27 +953,13 @@ jsonlib_dump (PyObject *self, PyObject *args)
 	base->ascii_only = ascii_only;
 	base->coerce_keys = coerce_keys;
 	
-	serializer.ascii_safe_encoding = ascii_safe;
-	
 	/* Implementation pointers */
 	base->module = PyModule_GetState (self);
-	base->append_ascii = stream_serializer_append_ascii;
-	base->append_unicode = stream_serializer_append_unicode;
-	
-	/* Small buffer for more efficient streaming */
-	Py_UNICODE stream_buffer[STREAM_BUFFER_SIZE];
-	serializer.buffer = stream_buffer;
-	serializer.buffer_size = 0;
+	base->flush = stream_serializer_flush;
 	
 	if (!serializer_run (base, value))
 	{ return NULL; }
 	
-	if (serializer.buffer_size > 0)
-	{
-		if (!stream_serializer_flush (&serializer))
-		{ return NULL; }
-	}
-	
 	Py_RETURN_NONE;
 }
 
@@ -995,15 +968,27 @@ serializer_run (Serializer *s, PyObject *value)
 {
 	const char *colon;
 	unsigned char retval;
+	Py_UNICODE fixed_buffer[FIXED_BUFFER_SIZE];
 	
 	/* Generate the colon constant */
 	colon = s->indent == Py_None? ":" : ": ";
 	if (!(s->colon = ascii_constant (colon)))
 	{ return 0; }
 	
+	/* Small buffer for more efficient streaming */
+	s->buffer = fixed_buffer;
+	s->buffer_size = 0;
+	
 	/* Run, clean up, return */
 	retval = serialize_object (s, value, 0, 0);
 	Py_DECREF (s->colon);
+	
+	if (s->buffer_size > 0)
+	{
+		if (!(s->flush(s)))
+		{ return 0; }
+	}
+	
 	return retval;
 }
 
@@ -1037,6 +1022,16 @@ serialize_object (Serializer *s, PyObject *value,
 		goto done;
 	}
 	
+	if (PyList_CheckExact (value))
+	{
+		if (!(iter = PyObject_GetIter (value)))
+		{ goto error; }
+		
+		retval = serialize_iterator (s, iter, value, indent_level);
+		Py_DECREF (iter);
+		goto done;
+	}
+	
 	/* Slow, general tests */
 	retval = serialize_atom (s, value);
 	if (retval == 0 || retval == 1)
@@ -1247,7 +1242,7 @@ serialize_mapping (Serializer *s, PyObject *mapping,
 	if (!serializer_separators (s, indent_level, &indent, &post_indent))
 	{ goto error; }
 	
-	if (!s->append_ascii (s, "{"))
+	if (!serializer_append_ascii (s, "{"))
 	{ goto error; }
 	
 	while ((serializer_get_mapping_pair (s, iter, &key, &value, &error)))
@@ -1255,16 +1250,16 @@ serialize_mapping (Serializer *s, PyObject *mapping,
 		if (first)
 		{ first = 0; }
 		
-		else if (!s->append_ascii (s, ","))
+		else if (!serializer_append_ascii (s, ","))
 		{ goto error; }
 		
-		if (indent && !s->append_unicode (s, indent))
+		if (indent && !serializer_append_unicode (s, indent))
 		{ goto error; }
 		
 		if (!serialize_object (s, key, indent_level + 1, 0))
 		{ goto error; }
 		
-		if (!s->append_unicode (s, s->colon))
+		if (!serializer_append_unicode (s, s->colon))
 		{ goto error; }
 		
 		if (!serialize_object (s, value, indent_level + 1, 0))
@@ -1276,10 +1271,10 @@ serialize_mapping (Serializer *s, PyObject *mapping,
 	if (error)
 	{ goto error; }
 	
-	if (post_indent &&!s->append_unicode (s, post_indent))
+	if (post_indent &&!serializer_append_unicode (s, post_indent))
 	{ goto error; }
 	
-	if (!s->append_ascii (s, "}"))
+	if (!serializer_append_ascii (s, "}"))
 	{ goto error; }
 	
 	goto success;
@@ -1309,7 +1304,7 @@ serialize_iterator (Serializer *s, PyObject *iter,
 	if (!serializer_separators (s, indent_level, &indent, &post_indent))
 	{ goto error; }
 	
-	if (!s->append_ascii (s, "["))
+	if (!serializer_append_ascii (s, "["))
 	{ goto error; }
 	
 	while ((item = PyIter_Next (iter)))
@@ -1317,10 +1312,10 @@ serialize_iterator (Serializer *s, PyObject *iter,
 		if (first)
 		{ first = 0; }
 		
-		else if (!s->append_ascii (s, ","))
+		else if (!serializer_append_ascii (s, ","))
 		{ goto error; }
 		
-		if (indent && !s->append_unicode (s, indent))
+		if (indent && !serializer_append_unicode (s, indent))
 		{ goto error; }
 		
 		if (!serialize_object (s, item, indent_level + 1, 0))
@@ -1329,10 +1324,10 @@ serialize_iterator (Serializer *s, PyObject *iter,
 		Py_DECREF (item);
 	}
 	
-	if (post_indent && !s->append_unicode (s, post_indent))
+	if (post_indent && !serializer_append_unicode (s, post_indent))
 	{ goto error; }
 	
-	if (!s->append_ascii (s, "]"))
+	if (!serializer_append_ascii (s, "]"))
 	{ goto error; }
 	
 	goto success;
@@ -1353,13 +1348,13 @@ serialize_atom_fast (Serializer *s, PyObject *value)
 	unsigned char retval;
 	
 	if (value == Py_True)
-	{ return s->append_unicode (s, m->true_str); }
+	{ return serializer_append_unicode (s, m->true_str); }
 	
 	if (value == Py_False)
-	{ return s->append_unicode (s, m->false_str); }
+	{ return serializer_append_unicode (s, m->false_str); }
 	
 	if (value == Py_None)
-	{ return s->append_unicode (s, m->null_str); }
+	{ return serializer_append_unicode (s, m->null_str); }
 	
 	if (PyUnicode_CheckExact (value))
 	{ return serialize_string (s, value); }
@@ -1370,7 +1365,7 @@ serialize_atom_fast (Serializer *s, PyObject *value)
 		if (!(str = PyObject_Str (value)))
 		{ return 0; }
 		
-		retval = s->append_unicode (s, str);
+		retval = serializer_append_unicode (s, str);
 		Py_DECREF (str);
 		return retval;
 	}
@@ -1439,11 +1434,11 @@ serialize_string (Serializer *s, PyObject *value)
 	
 	if (safe)
 	{
-		if (!s->append_ascii (s, "\""))
+		if (!serializer_append_ascii (s, "\""))
 		{ goto error; }
-		if (!s->append_unicode (s, value))
+		if (!serializer_append_unicode (s, value))
 		{ goto error; }
-		if (!s->append_ascii (s, "\""))
+		if (!serializer_append_ascii (s, "\""))
 		{ goto error; }
 		
 		return 1;
@@ -1457,7 +1452,7 @@ serialize_string (Serializer *s, PyObject *value)
 	else
 	{ escaped = escape_string_unicode (value); }
 	
-	if (escaped && s->append_unicode (s, escaped))
+	if (escaped && serializer_append_unicode (s, escaped))
 	{ goto success; }
 	
 error:
@@ -1485,7 +1480,7 @@ serialize_atom (Serializer *s, PyObject *value)
 		if (!(str = PyObject_Str (value)))
 		{ return 0; }
 		
-		retval = s->append_unicode (s, str);
+		retval = serializer_append_unicode (s, str);
 		Py_DECREF (str);
 		return retval;
 	}
@@ -1527,7 +1522,7 @@ serialize_dict (Serializer *s, PyObject *dict, unsigned int indent_level)
 	if (!serializer_separators (s, indent_level, &indent, &post_indent))
 	{ goto error; }
 	
-	if (!s->append_ascii (s, "{"))
+	if (!serializer_append_ascii (s, "{"))
 	{ goto error; }
 	
 	while (PyDict_Next (dict, &dict_pos, &key, &value))
@@ -1541,16 +1536,16 @@ serialize_dict (Serializer *s, PyObject *dict, unsigned int indent_level)
 		if (first)
 		{ first = 0; }
 		
-		else if (!s->append_ascii (s, ","))
+		else if (!serializer_append_ascii (s, ","))
 		{ goto error; }
 		
-		if (indent && !s->append_unicode (s, indent))
+		if (indent && !serializer_append_unicode (s, indent))
 		{ goto error; }
 		
 		if (!serialize_object (s, key, indent_level + 1, 0))
 		{ goto error; }
 		
-		if (!s->append_unicode (s, s->colon))
+		if (!serializer_append_unicode (s, s->colon))
 		{ goto error; }
 		
 		if (!serialize_object (s, value, indent_level + 1, 0))
@@ -1562,10 +1557,10 @@ serialize_dict (Serializer *s, PyObject *dict, unsigned int indent_level)
 		value = NULL;
 	}
 	
-	if (post_indent && !s->append_unicode (s, post_indent))
+	if (post_indent && !serializer_append_unicode (s, post_indent))
 	{ goto error; }
 	
-	if (!s->append_ascii (s, "}"))
+	if (!serializer_append_ascii (s, "}"))
 	{ goto error; }
 	
 	goto success;
@@ -1830,7 +1825,7 @@ serialize_float (Serializer *s, PyObject *value)
 	if (!(repr = PyObject_Repr (value)))
 	{ return 0; }
 	
-	retval = s->append_unicode (s, repr);
+	retval = serializer_append_unicode (s, repr);
 	Py_DECREF (repr);
 	return retval;
 }
@@ -1853,7 +1848,7 @@ serialize_complex (Serializer *s, PyObject *value)
 	if (!repr)
 	{ return 0; }
 	
-	retval = s->append_unicode (s, repr);
+	retval = serializer_append_unicode (s, repr);
 	Py_DECREF (repr);
 	return retval;
 }
@@ -1880,7 +1875,7 @@ serialize_decimal (Serializer *s, PyObject *value)
 	if (error)
 	{ retval = serializer_raise (s, error); }
 	else
-	{ retval = s->append_unicode (s, str); }
+	{ retval = serializer_append_unicode (s, str); }
 	
 	Py_DECREF (str);
 	return retval;
@@ -1902,82 +1897,14 @@ serializer_raise_reserved_cp (Serializer *s, Py_UNICODE c)
 }
 
 static unsigned char
-buffer_serializer_append_ascii (Serializer *base, const char *text)
+serializer_append_ascii (Serializer *s, const char *text)
 {
-	BufferSerializer *s = (BufferSerializer *) base;
-	size_t ii;
-	size_t len = strlen (text);
-	
-	if (!buffer_serializer_resize (s, len))
-	{ return 0; }
-	
-	for (ii = 0; ii < len; ii++)
-	{ s->buffer[s->buffer_size++] = text[ii]; }
-	
-	return 1;
-}
-
-static unsigned char
-buffer_serializer_append_unicode (Serializer *base, PyObject *text)
-{
-	size_t len;
-	Py_UNICODE *raw;
-	BufferSerializer *s = (BufferSerializer *) base;
-	
-	raw = PyUnicode_AS_UNICODE (text);
-	len = PyUnicode_GET_SIZE (text);
-	
-	if (!buffer_serializer_resize (s, len))
-	{ return 0; }
-	
-	memcpy (s->buffer + s->buffer_size, raw,
-	        len * sizeof (Py_UNICODE));
-	s->buffer_size += len;
-	return 1;
-}
-
-static unsigned char
-buffer_serializer_resize (BufferSerializer *s, size_t delta)
-{
-	size_t new_size;
-	Py_UNICODE *new_buf;
-	
-	new_size = s->buffer_size + delta;
-	if (s->buffer_max_size >= new_size)
-	{ return 1; }
-	
-	if (!s->buffer)
-	{
-		new_size = (delta > INITIAL_BUFFER_SIZE? delta : INITIAL_BUFFER_SIZE);
-		new_size = next_power_2 (1, new_size);
-		s->buffer = PyMem_Malloc (sizeof (Py_UNICODE) * new_size);
-		s->buffer_max_size = new_size;
-		return 1;
-	}
-	
-	new_size = next_power_2 (s->buffer_max_size, new_size);
-	new_buf = PyMem_Realloc (s->buffer, sizeof (Py_UNICODE) * new_size);
-	if (!new_buf)
-	{
-		PyMem_Free (s->buffer);
-		return 0;
-	}
-	s->buffer = new_buf;
-	s->buffer_max_size = new_size;
-	return 1;
-}
-
-static unsigned char
-stream_serializer_append_ascii (Serializer *base, const char *text)
-{
-	StreamSerializer *s = (StreamSerializer *) (base);
-	
 	/* Copy text to s->buffer, flushing as needed */
 	for (; *text; text++)
 	{
-		if (s->buffer_size == STREAM_BUFFER_SIZE)
+		if (s->buffer_size == FIXED_BUFFER_SIZE)
 		{
-			if (!stream_serializer_flush (s))
+			if (!(s->flush (s)))
 			{ return 0; }
 		}
 		s->buffer[s->buffer_size++] = *text;
@@ -1987,18 +1914,17 @@ stream_serializer_append_ascii (Serializer *base, const char *text)
 }
 
 static unsigned char
-stream_serializer_append_unicode (Serializer *base, PyObject *obj)
+serializer_append_unicode (Serializer *s, PyObject *obj)
 {
-	StreamSerializer *s = (StreamSerializer *) (base);
 	Py_UNICODE *text = PyUnicode_AS_UNICODE (obj);
 	Py_ssize_t ii, len_text = PyUnicode_GET_SIZE (obj);
 	
 	/* Copy text to s->buffer, flushing as needed */
 	for (ii = 0; ii < len_text; ii++)
 	{
-		if (s->buffer_size == STREAM_BUFFER_SIZE)
+		if (s->buffer_size == FIXED_BUFFER_SIZE)
 		{
-			if (!stream_serializer_flush (s))
+			if (!(s->flush (s)))
 			{ return 0; }
 		}
 		s->buffer[s->buffer_size++] = text[ii];
@@ -2008,21 +1934,126 @@ stream_serializer_append_unicode (Serializer *base, PyObject *obj)
 }
 
 static unsigned char
-stream_serializer_flush (StreamSerializer *s)
+buffer_serializer_flush (Serializer *base)
 {
-	PyObject *obj, *temp;
+	BufferSerializer *s = (BufferSerializer *)base;
+	BufferList *lst = s->current_buffer;
 	
-	if (s->encoding)
+	assert (base->buffer_size <= FIXED_BUFFER_SIZE);
+	
+	/* On the first flush, buffer pointers will be empty */
+	if (lst)
 	{
-		obj = PyUnicode_Encode(s->buffer, s->buffer_size, s->encoding, "strict");
-		if (!obj)
-		{ return 0; }
+		lst->next = alloc_buffer_list();
+		lst = lst->next;
 	}
 	
 	else
 	{
-		obj = PyUnicode_FromUnicode(s->buffer, s->buffer_size);
+		lst = alloc_buffer_list();
+		s->all_buffers = lst;
 	}
+	if (!lst)
+	{ return 0; }
+	
+	s->current_buffer = lst;
+	lst->buffer_size = base->buffer_size;
+	memcpy (lst->buffer, base->buffer,
+	        base->buffer_size * sizeof (Py_UNICODE));
+	base->buffer_size = 0;
+	
+	return 1;
+}
+
+static BufferList *
+alloc_buffer_list (void)
+{
+	BufferList *lst;
+	if (!(lst = PyMem_Malloc (sizeof (BufferList))))
+	{ return NULL; }
+	
+	lst->buffer_size = 0;
+	lst->next = NULL;
+	lst->buffer = PyMem_Malloc (sizeof (Py_UNICODE) * FIXED_BUFFER_SIZE);
+	if (!lst->buffer)
+	{ return NULL; }
+	
+	return lst;
+}
+
+static PyObject *
+buffer_serializer_merge (BufferSerializer *s)
+{
+	size_t total = 0, offset = 0;
+	BufferList *lst;
+	Py_UNICODE *raw_merged;
+	PyObject *merged;
+	
+	lst = s->all_buffers;
+	while (lst)
+	{
+		total += lst->buffer_size;
+		lst = lst->next;
+	}
+	
+	if (!(raw_merged = PyMem_Malloc (sizeof (Py_UNICODE) * total)))
+	{ return NULL; }
+	
+	lst = s->all_buffers;
+	while (lst)
+	{
+		memcpy (raw_merged + offset, lst->buffer,
+		        lst->buffer_size * sizeof (Py_UNICODE));
+		offset += lst->buffer_size;
+		lst = lst->next;
+	}
+	
+	if (s->encoding)
+	{
+		merged = PyUnicode_Encode (raw_merged, total, s->encoding,
+		                           "strict");
+	}
+	
+	else
+	{
+		merged = PyUnicode_FromUnicode (raw_merged, total);
+	}
+	PyMem_Free (raw_merged);
+	
+	return merged;
+}
+
+static void
+buffer_serializer_free (BufferSerializer *s)
+{
+	BufferList *lst = s->all_buffers;
+	while (lst)
+	{
+		BufferList *next = lst->next;
+		PyMem_Free (lst->buffer);
+		PyMem_Free (lst);
+		lst = next;
+	}
+}
+
+static unsigned char
+stream_serializer_flush (Serializer *base)
+{
+	PyObject *obj, *temp;
+	StreamSerializer *s = (StreamSerializer*)base;
+	
+	if (s->encoding)
+	{
+		obj = PyUnicode_Encode (base->buffer, base->buffer_size,
+		                        s->encoding, "strict");
+	}
+	
+	else
+	{
+		obj = PyUnicode_FromUnicode(base->buffer, base->buffer_size);
+	}
+	if (!obj)
+	{ return 0; }
 	
 	temp = PyObject_CallMethod (s->stream, "write", "O", obj);
 	Py_DECREF (obj);
@@ -2030,7 +2061,7 @@ stream_serializer_flush (StreamSerializer *s)
 	{ return 0; }
 	
 	Py_DECREF (temp);
-	s->buffer_size = 0;
+	base->buffer_size = 0;
 	return 1;
 }
 
